import importlib.util
import os
from pathlib import Path
import sys
import tempfile

from .base_renderer import BaseRenderer
import requests
from zipfile import ZipFile
from io import BytesIO


class RendererImporter:
    def __init__(self, template_module_url: str):
        if template_module_url.startswith("http://") or template_module_url.startswith("https://"):
            if not self.__is_github_repository(template_module_url):
                raise ValueError("URL must be a GitHub repository")
        self.template_module_url = template_module_url

    def __enter__(self):
        self.temp_dir = tempfile.TemporaryDirectory()
        self.temp_dir_path = Path(self.temp_dir.name)

        if self.__is_github_repository(self.template_module_url):
            plugin_dir = self.__download_github_repo()
        else:
            plugin_dir = self.template_module_url

        instance = self.__find_renderer_implementation(plugin_dir)
        if instance is None:
            raise ValueError("No renderer found")

        return instance

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.temp_dir is not None:
            self.temp_dir.cleanup()

    def __is_github_repository(self, url: str) -> bool:
        return url.startswith("https://github.com/")

    def __copy_plugin(self) -> str:
        plugin_name = os.path.basename(self.template_module_url)
        dest_dir = os.path.join(self.temp_dir_path, plugin_name)
        os.makedirs(dest_dir, exist_ok=True)
        for root, _, files in os.walk(self.template_module_url):
            for file in files:
                src_path = os.path.join(root, file)
                dest_path = os.path.join(dest_dir, file)
                with (
                    open(src_path, "rb") as src_file,
                    open(dest_path, "wb") as dest_file,
                ):
                    dest_file.write(src_file.read())
        return dest_dir

    def __download_github_repo(self) -> str:
        response = requests.get(f"{self.template_module_url}/archive/refs/heads/main.zip")
        if response.status_code != 200:
            raise ValueError("Failed to download repository")
        with ZipFile(BytesIO(response.content)) as zip_file:
            zip_file.extractall(self.temp_dir_path)

        repo_name = os.path.basename(self.template_module_url)
        extracted_dir = os.path.join(self.temp_dir_path, f"{repo_name}-main")
        return extracted_dir

    def __find_renderer_implementation(self, module_dir: str) -> BaseRenderer | None:
        sys.path.insert(0, module_dir)
        instance: BaseRenderer | None = None
        gitignore_path = os.path.join(module_dir, ".gitignore")
        if os.path.exists(gitignore_path):
            gitignore_content = Path(gitignore_path).read_text()
            files_to_ignore = gitignore_content.split("\n")
        else:
            files_to_ignore = []
        for root, _, files in os.walk(module_dir, topdown=False):
            rel_path = os.path.relpath(root, module_dir)
            if rel_path.startswith("."):  # Ignore hidden directories
                continue

            if os.path.dirname(rel_path) in files_to_ignore:  # Ignore directories listed in .gitignore
                continue

            for file in files:
                if instance is not None:
                    break
                if file.endswith(".py"):
                    module_name = os.path.splitext(file)[0]
                    spec = importlib.util.spec_from_file_location(module_name, os.path.join(root, file))
                    if spec is None:
                        continue
                    module = importlib.util.module_from_spec(spec)
                    if spec.loader is None:
                        continue
                    spec.loader.exec_module(module)
                    for attr_name in dir(module):
                        attr = getattr(module, attr_name)
                        if isinstance(attr, type) and issubclass(attr, BaseRenderer) and attr is not BaseRenderer:
                            instance = attr()
                            break
        sys.path.pop(0)
        return instance



from logging import Logger
import docker
import os
import shutil


class DockerImageBuilder:
    DOCKERFILE_NAME = "Dockerfile"
    NGINX_CONFIG_NAME = "nginx.conf"

    def __init__(self, docker_host: str, logger: Logger):
        self.docker_host = docker_host
        self.logger = logger

    def build_docker_image(self, image_name: str, image_version: str, gallery_output_path: str):
        if not self.check_if_docker_is_running():
            raise Exception("Docker engine is not running")

        shutil.copy(
            os.path.join(os.path.dirname(__file__), self.DOCKERFILE_NAME),
            os.path.join(gallery_output_path, self.DOCKERFILE_NAME),
        )

        shutil.copy(
            os.path.join(os.path.dirname(__file__), self.NGINX_CONFIG_NAME),
            os.path.join(gallery_output_path, self.NGINX_CONFIG_NAME),
        )

        client = self.__create_client()

        self.logger.info(f"Building image {image_name}:{image_version}")
        image, _ = client.images.build(path=gallery_output_path, tag=f"{image_name}:{image_version}", nocache=True)
        self.logger.info(f"Image {image.short_id} built successfully")

    def check_if_docker_is_running(self):
        try:
            client = self.__create_client()
            client.ping()
            return True
        except Exception:
            return False

    def __create_client(self) -> docker.DockerClient:
        client = docker.DockerClient(base_url=self.docker_host, timeout=10)
        return client



from abc import ABC, abstractmethod


class BaseStorageProvider(ABC):
    @abstractmethod
    def upload_image(self, image_data: bytes, directory: str, image_name: str) -> str:
        pass

    @abstractmethod
    def file_exists(self, directory: str, image_name: str) -> str | None:
        pass

    @abstractmethod
    def base_url(self) -> str:
        pass


class BaseSourceDataProvider(ABC):
    @abstractmethod
    def list_images(self, folder: str) -> list:
        pass

    @abstractmethod
    def get_image_data(self, directory: str, image_name: str) -> bytes:
        pass



from azure.storage.blob import BlobServiceClient
from azure.core.exceptions import ResourceNotFoundError
from .base_provider import BaseSourceDataProvider, BaseStorageProvider
from ..model import AzureBlobStorageConfig


class AzureBlobSourceDataProvider(BaseSourceDataProvider):
    def __init__(self, config: AzureBlobStorageConfig):
        self.config = config
        if not self.config.connection_string:
            raise ValueError("Azure Storage connection string not found.")

        self.blob_service_client = BlobServiceClient.from_connection_string(self.config.connection_string)
        self.container_client = self.blob_service_client.get_container_client(self.config.container)
        self.base_url_value = self.container_client.url + "/"

    def list_images(self, folder: str) -> list:
        image_blobs = []
        for blob in self.container_client.list_blobs(folder):
            if blob.name.lower().endswith((".png", ".jpg", ".jpeg")) and "thumbnails/" not in blob.name.lower():
                image_blobs.append(blob.name)
        return image_blobs

    def get_image_data(self, directory: str, image_name: str) -> bytes:
        blob_client = self.container_client.get_blob_client(f"{directory}/{image_name}")
        return blob_client.download_blob().readall()


class AzureBlobStorageProvider(BaseStorageProvider):
    def __init__(self, config: AzureBlobStorageConfig):
        self.config = config
        if not self.config.connection_string:
            raise ValueError("Azure Storage connection string not found.")

        self.blob_service_client = BlobServiceClient.from_connection_string(self.config.connection_string, api_version="2025-05-05")
        self.container_client = self.blob_service_client.get_container_client(self.config.container)
        self.base_url_value = self.container_client.url + "/"

    def upload_image(self, image_data, directory, image_name) -> str:
        blob_name = f"{directory}/{image_name}"
        if self.__blob_exists(blob_name):
            return blob_name
        blob_client = self.container_client.get_blob_client(f"{directory}/{image_name}")
        blob_client.upload_blob(image_data, overwrite=True)
        return blob_name

    def file_exists(self, directory: str, image_name: str) -> str | None:
        blob_name = f"{directory}/{image_name}"
        if self.__blob_exists(blob_name):
            return blob_name
        return None

    def __blob_exists(self, blob_name: str) -> bool:
        try:
            self.container_client.get_blob_client(blob_name).get_blob_properties()
            return True
        except ResourceNotFoundError:
            return False
        except Exception:
            return False
        
    def base_url(self) -> str:
        return self.base_url_value



import logging
import os
import argparse

from dotenv import load_dotenv

from .model import GalleryConfig, LocalStorageConfig
from .config import load_config
from .storage import get_source_provider, get_storage_provider
from .renderers import RendererImporter
from .favicon import get_favicon_provider
from .image import ThumbnailCreator
from .gallery_generator import GalleryGenerator

load_dotenv()
logging.basicConfig(
    level=logging.INFO,
)
logger = logging.getLogger("ggallery")
logging.getLogger("azure.core").setLevel(logging.WARNING)


class Arguments:
    file: str


def main():
    parser = argparse.ArgumentParser(description="Generate a static HTML photo gallery.", prog="ggallery")

    if not os.path.exists("gallery.yaml"):
        parser.add_argument("-f", "--file", help="Path to the configuration file", required=True)
    else:
        parser.add_argument(
            "-f",
            "--file",
            help="Path to the configuration file",
            default="gallery.yaml",
        )

    args: Arguments = parser.parse_args(namespace=Arguments)  # type: ignore

    if not os.path.exists(args.file):
        print(f"Configuration file {args.file} not found.")
        return

    gallery: GalleryConfig = load_config(args.file)
    source_provider = get_source_provider(gallery.data_source)

    if isinstance(gallery.data_storage, LocalStorageConfig):
        gallery.data_storage.path = gallery.output.path

    target_provider = get_storage_provider(gallery.data_storage)
    favicon_provider = get_favicon_provider(gallery.favicon.type) if gallery.favicon is not None else None
    with RendererImporter(gallery.template.url) as renderer:
        thumbnail_creator = ThumbnailCreator()
        gallery_generator = GalleryGenerator(
            source_storage_provider=source_provider,
            target_storage_provider=target_provider,
            thumbnail_creator=thumbnail_creator,
            favicon_provider=favicon_provider,
            renderer=renderer,
            logger=logger,
        )

        gallery_generator.create_gallery(gallery)


if __name__ == "__main__":
    main()



from abc import abstractmethod, ABC
from dataclasses import dataclass
from ..model import FaviconConfig


@dataclass
class Favicon:
    url: str | None
    content: bytes | None
    file_type: str | None = None


class BaseFaviconProvider(ABC):
    @abstractmethod
    def get_favicon(self, config: FaviconConfig) -> Favicon:
        pass



from logging import Logger
import os
from pathlib import Path

from .docker import DockerImageBuilder
from .model import GalleryConfig, PhotoConfig, RendererParameters
from .storage import BaseStorageProvider, BaseSourceDataProvider
from .image import ThumbnailCreator
from .favicon import BaseFaviconProvider
from .renderers import BaseRenderer


class GalleryGenerator:
    def __init__(
        self,
        source_storage_provider: BaseSourceDataProvider,
        target_storage_provider: BaseStorageProvider,
        thumbnail_creator: ThumbnailCreator,
        favicon_provider: BaseFaviconProvider | None,
        renderer: BaseRenderer,
        logger: Logger,
    ) -> None:
        self.source_storage_provider = source_storage_provider
        self.target_storage_provider = target_storage_provider
        self.thumbnail_creator = thumbnail_creator
        self.favicon_provider = favicon_provider
        self.renderer = renderer
        self.logger = logger

    def create_gallery(self, gallery: GalleryConfig) -> None:
        albums = gallery.albums
        thumbnail_height = gallery.thumbnail.height
        actual_thumbnail_height = thumbnail_height * 2  # Double the height for better quality

        for album in albums:
            self.logger.info(f"Processing album {album.title}.")
            if album.source is None:
                continue

            images = self.source_storage_provider.list_images(album.source)
            photos: list[PhotoConfig] = []
            source_catalog = album.source
            target_catalog = album.source

            for image_name in images:
                thumbnail_name = self.thumbnail_creator.create_thumbnail_name(image_name, actual_thumbnail_height)

                image_uri = self.target_storage_provider.file_exists(target_catalog, image_name)
                thumbnail_uri = self.target_storage_provider.file_exists(target_catalog, thumbnail_name)

                if not thumbnail_uri or not image_uri:
                    image = self.source_storage_provider.get_image_data(source_catalog, image_name)

                    if not image_uri:
                        self.logger.info(f"Uploading image {image_name} to storage.")
                        image_uri = self.target_storage_provider.upload_image(image, target_catalog, image_name)

                    if not thumbnail_uri:
                        thumbnail = self.thumbnail_creator.create_thumbnail(image, actual_thumbnail_height)
                        self.logger.info(f"Uploading thumbnail {thumbnail_name} to storage.")
                        thumbnail_uri = self.target_storage_provider.upload_image(
                            thumbnail, target_catalog, thumbnail_name
                        )

                photo = PhotoConfig(filename=image_uri, thumbnail=thumbnail_uri)
                if album.photos is not None:
                    existing_photo_metadata = next((p for p in album.photos if p.source == image_name), None)
                    if existing_photo_metadata:
                        photo = existing_photo_metadata
                        photo.filename = image_uri
                        photo.thumbnail = thumbnail_uri

                photos.append(photo)

            if album.cover is not None:
                thumbnail_name = self.thumbnail_creator.create_thumbnail_name(album.cover, actual_thumbnail_height)
                album.cover = self.target_storage_provider.file_exists(target_catalog, thumbnail_name)

            album.photos = photos

        output_config = gallery.output
        favicon_url = self.__get_favicon_url(gallery)

        render_parameters = RendererParameters(
            albums=albums,
            base_url=self.target_storage_provider.base_url(),
            thumbnail_height=thumbnail_height,
            title=gallery.title,
            subtitle=gallery.subtitle,
            favicon=favicon_url,
            template_parameters=gallery.template.parameters,
        )

        self.logger.info("Rendering gallery website.")
        rendered_files = self.renderer.render(render_parameters)
        if not isinstance(rendered_files, list):
            rendered_files = [rendered_files]
        for rendered_file in rendered_files:
            self.__write_content_to_output_directory(output_config.path, rendered_file.name, rendered_file.content)

        if gallery.docker is not None:
            docker_image_builder = DockerImageBuilder(gallery.docker.host, self.logger)
            docker_image_builder.build_docker_image(
                gallery.docker.image_name,
                gallery.docker.image_version,
                output_config.path,
            )

    def __write_content_to_output_directory(self, output_path: str, file_name: str, content: bytes | str) -> None:
        output_directory = Path(output_path)
        file_path = output_directory / file_name
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        if isinstance(content, str):
            with file_path.open("w") as f:
                f.write(content)
        elif isinstance(content, bytes):
            with file_path.open("wb") as f:
                f.write(content)

    def __get_favicon_url(self, gallery: GalleryConfig) -> str | None:
        favicon_config = gallery.favicon
        output_config = gallery.output
        favicon_url = None
        if favicon_config is not None and self.favicon_provider is not None:
            favicon = self.favicon_provider.get_favicon(favicon_config)
            if favicon.url is not None:
                favicon_url = favicon.url
            elif favicon.content is not None and favicon.file_type is not None:
                favicon_name = f"favicon.{favicon.file_type}"
                self.__write_content_to_output_directory(output_config.path, favicon_name, favicon.content)
                favicon_url = favicon_name
        return favicon_url



from ..model import FaviconConfig
from .base_favicon_provider import BaseFaviconProvider, Favicon


class LocalFaviconProvider(BaseFaviconProvider):
    def get_favicon(self, config: FaviconConfig) -> Favicon:
        with open(config.name, "rb") as file:
            file_extension = config.name.split(".")[-1]
            return Favicon(url=None, content=file.read(), file_type=file_extension)



from pathlib import Path
from unittest.mock import MagicMock
from ggallery.gallery_generator import GalleryGenerator
from ggallery.model import AlbumConfig, PhotoConfig, LocalStorageConfig, AzureBlobStorageConfig, RenderedFile, RendererParameters, OutputConfig
from ggallery.storage import get_source_provider, get_storage_provider
from ggallery.image import ThumbnailCreator
from ggallery.renderers import BaseRenderer
from azure.storage.blob import BlobServiceClient
from testcontainers.azurite import AzuriteContainer
from tests.gallery_config_factory import GalleryConfigFactory
import logging
import requests

logging.basicConfig(
    level=logging.INFO,
)
logger = logging.getLogger("ggallery-tests")
logging.getLogger("azure.core").setLevel(logging.WARNING)


class TestGalleryGenerator:
    blob_container_name = "photos"

    def setup_method(self):
        self.thumbnail_creator = MagicMock()
        self.favicon_provider = MagicMock()
        self.logger = MagicMock()
        self.azurite_container = AzuriteContainer()
        self.azurite_container.start()
        client = BlobServiceClient.from_connection_string(
            self.azurite_container.get_connection_string(),
            api_version="2025-05-05"
        )
        client.create_container(self.blob_container_name, public_access="blob")

    def teardown_method(self):
        self.azurite_container.stop()

    def test_create_gallery_with_local_source_and_azure_target(self):
        script_directory = Path(__file__).parent
        assets_path = script_directory / "assets"

        gallery_config = GalleryConfigFactory().with_albums([
            AlbumConfig(
                title="test album",
                subtitle="test subtitle",
                source="album",
                cover="peter-herrmann-4cWu6XpOe58-unsplash.jpg",
                photos=[
                    PhotoConfig(
                        title="Peter Herrmann Photo",
                        filename="peter-herrmann-4cWu6XpOe58-unsplash.jpg",
                        source="peter-herrmann-4cWu6XpOe58-unsplash.jpg",
                    ),
                    PhotoConfig(
                        title="Peter Thomas Photo",
                        filename="peter-thomas-ep-6kGyBSXk-unsplash.jpg",
                        source="peter-thomas-ep-6kGyBSXk-unsplash.jpg",
                    ),
                    PhotoConfig(
                        title="Victor Rosario Photo",
                        filename="victor-rosario-f2SquFjpESg-unsplash.jpg",
                        source="victor-rosario-f2SquFjpESg-unsplash.jpg",
                    )
                ]
            )
        ]).with_data_source(
            LocalStorageConfig(path=assets_path.absolute().as_posix())
        ).with_data_storage(
            AzureBlobStorageConfig(container=self.blob_container_name, connection_string=self.azurite_container.get_connection_string())
        ).with_title(
            "Test Gallery"
        ).with_subtitle(
            "Test Subtitle"
        ).with_output(
            OutputConfig(
                path=assets_path.absolute().as_posix()
            )
        ).build()

        source_provider = get_source_provider(gallery_config.data_source)
        target_provider = get_storage_provider(gallery_config.data_storage)
        thumbnail_creator = ThumbnailCreator()
        renderer = RendererStub()

        gallery_generator = GalleryGenerator(
            source_storage_provider=source_provider,
            target_storage_provider=target_provider,
            thumbnail_creator=thumbnail_creator,
            favicon_provider=None,
            renderer=renderer,
            logger=logger,
        )

        gallery_generator.create_gallery(gallery_config)

        expected_export_file = assets_path / renderer.rendered_gallery_name
        assert expected_export_file.exists(), f"Expected export file {expected_export_file} does not exist"

        export_file_content = expected_export_file.read_text(encoding="utf-8")
        rendered_parameters = RendererParameters.model_validate_json(export_file_content)
        assert rendered_parameters.title == gallery_config.title
        assert rendered_parameters.subtitle == gallery_config.subtitle
        assert len(rendered_parameters.albums) == 1
        rendered_album = rendered_parameters.albums[0]
        config_album = gallery_config.albums[0]
        assert rendered_album.title == config_album.title
        assert rendered_album.subtitle == config_album.subtitle
        assert rendered_album.photos is not None
        assert config_album.photos is not None

        for rendered_photo, config_photo in zip(rendered_album.photos, config_album.photos):
            assert rendered_photo.filename is not None
            assert config_photo.filename is not None
            rendered_photo_url = rendered_parameters.base_url + rendered_photo.filename
            source_photo_path = assets_path / config_photo.filename

            rendered_photo_content = requests.get(rendered_photo_url).content
            source_photo_content = source_photo_path.read_bytes()
            assert rendered_photo_content == source_photo_content

        expected_export_file.unlink() 


class RendererStub(BaseRenderer):
    rendered_gallery_name = "rendered_gallery.json"

    def render(self, parameters: RendererParameters) -> list[RenderedFile] | RenderedFile:
        return RenderedFile(
            name=self.rendered_gallery_name, 
            content=parameters.model_dump_json(indent=4).encode("utf-8")
        )


from .base_favicon_provider import BaseFaviconProvider
from .fontawesome_favicon_provider import FontAwesomeFaviconProvider
from .local_favicon_provider import LocalFaviconProvider


def get_favicon_provider(name: str) -> BaseFaviconProvider:
    if name == "fontawesome":
        return FontAwesomeFaviconProvider()
    elif name == "local":
        return LocalFaviconProvider()
    else:
        raise ValueError(f"Unsupported favicon provider: {name}")



from ggallery.model import AlbumConfig, GalleryConfig, LocalStorageConfig, OutputConfig, TemplateConfig, ThumbnailsConfig


class GalleryConfigFactory:
    def __init__(self):
        self._albums = [
            AlbumConfig(
                title="Test Album",
                source="source",
                photos=None,
                cover=None
            )
        ]
        self._thumbnail = ThumbnailsConfig(height=100)
        self._output = OutputConfig(path="output")
        self._template = TemplateConfig(parameters={}, url="http://example.com/template")
        self._favicon = None
        self._docker = None
        self._title = "Test Gallery"
        self._subtitle = "Test Subtitle"
        self._data_source = LocalStorageConfig(path="source", type="local")
        self._data_storage = LocalStorageConfig(path="source", type="local")

    def with_albums(self, albums):
        self._albums = albums
        return self

    def with_thumbnail(self, thumbnail):
        self._thumbnail = thumbnail
        return self

    def with_output(self, output):
        self._output = output
        return self

    def with_template(self, template):
        self._template = template
        return self

    def with_favicon(self, favicon):
        self._favicon = favicon
        return self

    def with_docker(self, docker):
        self._docker = docker
        return self

    def with_title(self, title):
        self._title = title
        return self

    def with_subtitle(self, subtitle):
        self._subtitle = subtitle
        return self

    def with_data_source(self, data_source):
        self._data_source = data_source
        return self

    def with_data_storage(self, data_storage):
        self._data_storage = data_storage
        return self

    def build(self):
        return GalleryConfig(
            albums=self._albums,
            thumbnail=self._thumbnail,
            output=self._output,
            template=self._template,
            favicon=self._favicon,
            docker=self._docker,
            title=self._title,
            subtitle=self._subtitle,
            data_source=self._data_source,
            data_storage=self._data_storage
        )


from ..model import FaviconConfig
from .base_favicon_provider import BaseFaviconProvider, Favicon

import requests


class FontAwesomeFaviconProvider(BaseFaviconProvider):
    def get_favicon(self, config: FaviconConfig) -> Favicon:
        url = f"https://cdn.jsdelivr.net/npm/@fortawesome/fontawesome-free@6/svgs/solid/{config.name}.svg"
        # fetch the content of the favicon
        response = requests.get(url)
        response.raise_for_status()
        return Favicon(url=None, content=response.content, file_type="svg")

        return Favicon(
            url=f"https://cdn.jsdelivr.net/npm/@fortawesome/fontawesome-free@6/svgs/solid/{config.name}.svg",
            content=None,
        )
